//=============================================================================
// AssignExprs.g - Recognize Assignment Statements and Arithmetic Expressions
//
// Author:  Bary W Pollack
// Date:    Feb. 14, 2006
//
//=============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
}


//-----------------------------------------------------------------------------
// Define a Parser, call it AssignExprsRecognizer
//-----------------------------------------------------------------------------

class AssignExprsRecognizer extends Parser;

options {
    defaultErrorHandler = false;    // Don't generate parser error handlers
}


// Define some methods and variables to use in the generated parser
{
    // Define a main
    public static void main(String[] args) {
        // Use a try/catch block for parser exceptions
        try {
            // if we have at least one command-line argument
            if (args.length > 0 ) {
                System.out.println();
                System.out.println("AssignExprs Parsing...");
                System.out.println();

                // For each directory/file specified on the command line
                for (int i = 0; i < args.length; i++)
                    doFile(new File(args[i]));  // parse it
            } else
                System.out.println("Usage: java AssignExprsRecogizer " +
                                   "[ file | <directory name> ]...");

        } catch (Exception e) {
            System.out.println("Exception: " + e);
            e.printStackTrace(System.out);   // so we can get a stack trace
        }
    }


    // This method decides what action to take based on the
    // type of file we are looking at -- directory or source file
    // It is designed to process a sequence of files/directories
    public static void doFile(File f) throws Exception {
        // If this is a directory, walk each file/dir in that directory
        if (f.isDirectory()) {
            String files[] = f.list();
            for (int i = 0; i < files.length; i++)
                doFile(new File(f, files[i]));        
        // otherwise, if this is a source file, parse it
        } else if ((f.getName().length() >= 5) &&
                   f.getName().substring(f.getName().length()-4).equals(".txt")) {
            //System.out.println("  File: " + f.getAbsolutePath());  //FOO
            System.out.println("  File: " + f.getName());
            parseFile(new FileInputStream(f));
        }
    }

    // Here's where we do the real work...  parse an entire file
    public static void parseFile(InputStream s) throws Exception {
        System.out.println();
        try {
            // Create a scanner that reads from the input stream passed to us
            AssignExprsLexer lexer = new AssignExprsLexer(s);

            // Create a parser that reads from the scanner
            AssignExprsRecognizer parser = new AssignExprsRecognizer(lexer);

            // Start parsing at the program rule
            pSymbols = new TreeMap<String, Symbol>();   // initialize the symbol table
            parser.program();
            System.out.println("* Success *");

        } catch (Exception e) {
            System.out.println("Parser exception: " + e);
            System.out.println();
            e.printStackTrace();   // so we can get a stack trace

        } finally {
            System.out.println();
            if (false) {           // control display of symbol table  //FOO
                displaySymbols();
                System.out.println();
            }
        }
    }

    // An extremely tiny symbol table utility
    // Contains symbol, its type (kind), source line number where last used

    public enum Kind { IDENTIFIER, INTEGER };

    public class Symbol {
        Symbol(String sName, Kind eKind, int nLno) {
            this.sName = sName;  this.eKind = eKind;  this.nLno = nLno;
        }
        public String toString() {
            return "<" + sName + " " + eKind + " " + nLno + ">";
        }
        private String sName;
        private Kind   eKind;
        private int    nLno;
    }

    public static TreeMap<String, Symbol> pSymbols;  // this is the symbol table

    static void displaySymbols() {
        System.out.println("Symbol Table (" + pSymbols.size() + " entries)");
        int n = 1;
        Set set = pSymbols.entrySet();
        for (Iterator i = set.iterator(); i.hasNext(); n++) {
            Map.Entry e = (Map.Entry) i.next();
            String s = ((String)e.getKey()) + "        ";
            System.out.print("   " + (n + "    ").substring(0,3));
            System.out.print("  " + s.substring(0, 8) + " ");
            System.out.println((Symbol)e.getValue());
        }
    }

}

// ==================================================================
// Grammar for a collection of statements containing assignment
// statements, arithmetic expressions, and procedure calls
// ==================================================================

program
  : programBody 
    // end-of-file
  ;

programBody
  : "begin"
       statementList
    "end"
  ;

constantValue
  : intlit:INTLIT
      { System.out.println("intlit=" + intlit.getText());    //FOO
        pSymbols.put(intlit.getText(), 
                     new Symbol(intlit.getText(), Kind.INTEGER, intlit.getLine()));
      }
  ;

statement
  : (IDENT (LPAREN|SEMI))=> procedureCallStatement
  | assignmentStatement
  ;

statementList
  : statement statementList
  | // nothing
  ;

assignmentStatement
  : variableReference EQUALS expression SEMI
  ;

procedureCallStatement
  : ident:IDENT (actualParameters)? SEMI
      { System.out.println("ident= " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), Kind.IDENTIFIER, ident.getLine()));
      }
  ;

actualParameters
  : LPAREN (expression (COMMA expression)*)? RPAREN
  ;

procedureCall
  : ident:IDENT actualParameters
      { System.out.println("ident= " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), Kind.IDENTIFIER, ident.getLine()));
      }
  ;

variableReference
  : ident:IDENT
      { System.out.println("ident= " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), Kind.IDENTIFIER, ident.getLine()));
      }
    ( LBRACKET expression RBRACKET )*
  ;

primitiveElement
  : (IDENT LPAREN)=> procedureCall
  | variableReference
  | constantValue
  | LPAREN expression RPAREN
  ;

signedExpression
  : ((PLUS|MINUS))* primitiveElement
  ;

multiplyingExpression
  : signedExpression ((TIMES|DIV) signedExpression)*
  ;

expression
  : multiplyingExpression ((PLUS|MINUS) multiplyingExpression)*
  ;

//----------------------------------------------------------------------------
// The AssignExprs scanner - recognize literals and identifiers (variables)
//----------------------------------------------------------------------------

class AssignExprsLexer extends Lexer;

options {
    charVocabulary = '\0'..'\377';
    testLiterals = false;        // don't automatically test for literals
    k = 2;                       // two characters of lookahead
}


// Single-line comments
COMMENT
  : "//" (~('\n'|'\r'))*
    { $setType(Token.SKIP); }
  ;


// Literals
protected DIGIT : '0'..'9';

INTLIT : (DIGIT)+ ;


// Keywords are literals in the parser grammar

// Operators
SEMI     : ';' ;
COMMA    : ',' ;
EQUALS   : '=' ;
LBRACKET : '[' ;
RBRACKET : ']' ;
LPAREN   : '(' ;
RPAREN   : ')' ;
PLUS     : '+' ;
MINUS    : '-' ;
TIMES    : '*' ;
DIV      : '/' ;


// Whitespace -- ignored
WS
  : ( ' '
    | '\t'
    | '\f'

    // Handle newlines
    | ( "\r\n"  // Evil DOS
      | '\r'    // Macintosh
      | '\n'    // Unix (the right way)
      )
      { newline(); }
    )
    { $setType(Token.SKIP); }
  ;


// An identifier.  Note that testLiterals is set to true.  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
  options { testLiterals=true; }
  : ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9')*
  ;

// EOF
